from __future__ import annotations

import datetime
import warnings
from collections.abc import Sequence
from typing import Literal, overload

import pandas as pd
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import Datapoints, DatapointsArrayList, DatapointsList, TimeSeriesList
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList
from cognite.client.data_classes.datapoints import Aggregate

from cognite.powerops.client._generated.data_classes import (
    Generator,
    GeneratorApply,
    GeneratorApplyList,
    GeneratorFields,
    GeneratorList,
    GeneratorTextFields,
)
from cognite.powerops.client._generated.data_classes._generator import _GENERATOR_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, INSTANCE_QUERY_LIMIT, Aggregations, TypeAPI

ColumnNames = Literal[
    "name",
    "pMin",
    "penstock",
    "startcost",
    "startStopCost",
    "isAvailableTimeSeries",
    "generatorEfficiencyCurve",
    "turbineEfficiencyCurve",
]


class GeneratorStartStopCostQuery:
    def __init__(
        self,
        client: CogniteClient,
        view_id: dm.ViewId,
        timeseries_limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ):
        self._client = client
        self._view_id = view_id
        self._timeseries_limit = timeseries_limit
        self._filter = filter

    def retrieve(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsList([])

    def retrieve_arrays(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsArrayList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_arrays(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsArrayList([])

    def retrieve_dataframe(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "startStopCost",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_dataframe_in_tz(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "startStopCost",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe_in_tz(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_latest(
        self,
        before: None | int | str | datetime.datetime = None,
    ) -> Datapoints | DatapointsList | None:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_latest(
                external_id=list(external_ids),
                before=before,
            )
        else:
            return None

    def plot(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "startStopCost",
        warning: bool = True,
        **kwargs,
    ) -> None:
        if warning:
            warnings.warn(
                "This methods if an experiment and might be removed in the future without notice.", stacklevel=2
            )
        if all(isinstance(time, datetime.datetime) and time.tzinfo is not None for time in [start, end]):
            df = self.retrieve_dataframe_in_tz(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        else:
            df = self.retrieve_dataframe(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        df.plot(**kwargs)

    def _retrieve_timeseries_external_ids_with_extra(
        self, extra_properties: ColumnNames | list[ColumnNames] = "startStopCost"
    ) -> dict[str, list[str]]:
        return _retrieve_timeseries_external_ids_with_extra_start_stop_cost(
            self._client,
            self._view_id,
            self._filter,
            self._timeseries_limit,
            extra_properties,
        )

    @staticmethod
    def _rename_columns(
        external_ids: dict[str, list[str]],
        df: pd.DataFrame,
        column_names: ColumnNames | list[ColumnNames],
        include_aggregate_name: bool,
        include_granularity_name: bool,
    ) -> pd.DataFrame:
        if isinstance(column_names, str) and column_names == "startStopCost":
            return df
        splits = sum(included for included in [include_aggregate_name, include_granularity_name])
        if splits == 0:
            df.columns = ["-".join(external_ids[external_id]) for external_id in df.columns]
        else:
            column_parts = (col.rsplit("|", maxsplit=splits) for col in df.columns)
            df.columns = [
                "-".join(external_ids[external_id]) + "|" + "|".join(parts) for external_id, *parts in column_parts
            ]
        return df


class GeneratorStartStopCostAPI:
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        self._client = client
        self._view_id = view_id

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> GeneratorStartStopCostQuery:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )

        return GeneratorStartStopCostQuery(
            client=self._client,
            view_id=self._view_id,
            timeseries_limit=limit,
            filter=filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> TimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )
        external_ids = _retrieve_timeseries_external_ids_with_extra_start_stop_cost(
            self._client, self._view_id, filter_, limit
        )
        if external_ids:
            return self._client.time_series.retrieve_multiple(external_ids=list(external_ids))
        else:
            return TimeSeriesList([])


def _retrieve_timeseries_external_ids_with_extra_start_stop_cost(
    client: CogniteClient,
    view_id: dm.ViewId,
    filter_: dm.Filter | None,
    limit: int,
    extra_properties: ColumnNames | list[ColumnNames] = "startStopCost",
) -> dict[str, list[str]]:
    properties = ["startStopCost"]
    if extra_properties == "startStopCost":
        ...
    elif isinstance(extra_properties, str) and extra_properties != "startStopCost":
        properties.append(extra_properties)
    elif isinstance(extra_properties, list):
        properties.extend([prop for prop in extra_properties if prop != "startStopCost"])
    else:
        raise ValueError(f"Invalid value for extra_properties: {extra_properties}")

    if isinstance(extra_properties, str):
        extra_list = [extra_properties]
    else:
        extra_list = extra_properties
    has_data = dm.filters.HasData([dm.ContainerId("power-ops", "Generator")], [view_id])
    filter_ = dm.filters.And(filter_, has_data) if filter_ else has_data

    cursor = None
    external_ids: dict[str, list[str]] = {}
    total_retrieved = 0
    while True:
        query_limit = min(INSTANCE_QUERY_LIMIT, limit - total_retrieved)
        selected_nodes = dm.query.NodeResultSetExpression(filter=filter_, limit=query_limit)
        query = dm.query.Query(
            with_={
                "nodes": selected_nodes,
            },
            select={
                "nodes": dm.query.Select(
                    [dm.query.SourceSelector(view_id, properties)],
                )
            },
            cursors={"nodes": cursor},
        )
        result = client.data_modeling.instances.query(query)
        batch_external_ids = {
            node.properties[view_id]["startStopCost"]: [node.properties[view_id].get(prop, "") for prop in extra_list]
            for node in result.data["nodes"].data
        }
        total_retrieved += len(batch_external_ids)
        external_ids.update(batch_external_ids)
        cursor = result.cursors["nodes"]
        if total_retrieved >= limit or cursor is None:
            break
    return external_ids


class GeneratorIsAvailableTimeSeriesQuery:
    def __init__(
        self,
        client: CogniteClient,
        view_id: dm.ViewId,
        timeseries_limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ):
        self._client = client
        self._view_id = view_id
        self._timeseries_limit = timeseries_limit
        self._filter = filter

    def retrieve(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsList([])

    def retrieve_arrays(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsArrayList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_arrays(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsArrayList([])

    def retrieve_dataframe(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "isAvailableTimeSeries",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_dataframe_in_tz(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "isAvailableTimeSeries",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe_in_tz(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_latest(
        self,
        before: None | int | str | datetime.datetime = None,
    ) -> Datapoints | DatapointsList | None:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_latest(
                external_id=list(external_ids),
                before=before,
            )
        else:
            return None

    def plot(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "isAvailableTimeSeries",
        warning: bool = True,
        **kwargs,
    ) -> None:
        if warning:
            warnings.warn(
                "This methods if an experiment and might be removed in the future without notice.", stacklevel=2
            )
        if all(isinstance(time, datetime.datetime) and time.tzinfo is not None for time in [start, end]):
            df = self.retrieve_dataframe_in_tz(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        else:
            df = self.retrieve_dataframe(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        df.plot(**kwargs)

    def _retrieve_timeseries_external_ids_with_extra(
        self, extra_properties: ColumnNames | list[ColumnNames] = "isAvailableTimeSeries"
    ) -> dict[str, list[str]]:
        return _retrieve_timeseries_external_ids_with_extra_is_available_time_series(
            self._client,
            self._view_id,
            self._filter,
            self._timeseries_limit,
            extra_properties,
        )

    @staticmethod
    def _rename_columns(
        external_ids: dict[str, list[str]],
        df: pd.DataFrame,
        column_names: ColumnNames | list[ColumnNames],
        include_aggregate_name: bool,
        include_granularity_name: bool,
    ) -> pd.DataFrame:
        if isinstance(column_names, str) and column_names == "isAvailableTimeSeries":
            return df
        splits = sum(included for included in [include_aggregate_name, include_granularity_name])
        if splits == 0:
            df.columns = ["-".join(external_ids[external_id]) for external_id in df.columns]
        else:
            column_parts = (col.rsplit("|", maxsplit=splits) for col in df.columns)
            df.columns = [
                "-".join(external_ids[external_id]) + "|" + "|".join(parts) for external_id, *parts in column_parts
            ]
        return df


class GeneratorIsAvailableTimeSeriesAPI:
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        self._client = client
        self._view_id = view_id

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> GeneratorIsAvailableTimeSeriesQuery:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )

        return GeneratorIsAvailableTimeSeriesQuery(
            client=self._client,
            view_id=self._view_id,
            timeseries_limit=limit,
            filter=filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> TimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )
        external_ids = _retrieve_timeseries_external_ids_with_extra_is_available_time_series(
            self._client, self._view_id, filter_, limit
        )
        if external_ids:
            return self._client.time_series.retrieve_multiple(external_ids=list(external_ids))
        else:
            return TimeSeriesList([])


def _retrieve_timeseries_external_ids_with_extra_is_available_time_series(
    client: CogniteClient,
    view_id: dm.ViewId,
    filter_: dm.Filter | None,
    limit: int,
    extra_properties: ColumnNames | list[ColumnNames] = "isAvailableTimeSeries",
) -> dict[str, list[str]]:
    properties = ["isAvailableTimeSeries"]
    if extra_properties == "isAvailableTimeSeries":
        ...
    elif isinstance(extra_properties, str) and extra_properties != "isAvailableTimeSeries":
        properties.append(extra_properties)
    elif isinstance(extra_properties, list):
        properties.extend([prop for prop in extra_properties if prop != "isAvailableTimeSeries"])
    else:
        raise ValueError(f"Invalid value for extra_properties: {extra_properties}")

    if isinstance(extra_properties, str):
        extra_list = [extra_properties]
    else:
        extra_list = extra_properties
    has_data = dm.filters.HasData([dm.ContainerId("power-ops", "Generator")], [view_id])
    filter_ = dm.filters.And(filter_, has_data) if filter_ else has_data

    cursor = None
    external_ids: dict[str, list[str]] = {}
    total_retrieved = 0
    while True:
        query_limit = min(INSTANCE_QUERY_LIMIT, limit - total_retrieved)
        selected_nodes = dm.query.NodeResultSetExpression(filter=filter_, limit=query_limit)
        query = dm.query.Query(
            with_={
                "nodes": selected_nodes,
            },
            select={
                "nodes": dm.query.Select(
                    [dm.query.SourceSelector(view_id, properties)],
                )
            },
            cursors={"nodes": cursor},
        )
        result = client.data_modeling.instances.query(query)
        batch_external_ids = {
            node.properties[view_id]["isAvailableTimeSeries"]: [
                node.properties[view_id].get(prop, "") for prop in extra_list
            ]
            for node in result.data["nodes"].data
        }
        total_retrieved += len(batch_external_ids)
        external_ids.update(batch_external_ids)
        cursor = result.cursors["nodes"]
        if total_retrieved >= limit or cursor is None:
            break
    return external_ids


class GeneratorAPI(TypeAPI[Generator, GeneratorApply, GeneratorList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Generator,
            class_apply_type=GeneratorApply,
            class_list=GeneratorList,
        )
        self._view_id = view_id
        self.start_stop_cost = GeneratorStartStopCostAPI(client, view_id)
        self.is_available_time_series = GeneratorIsAvailableTimeSeriesAPI(client, view_id)

    def apply(
        self, generator: GeneratorApply | Sequence[GeneratorApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(generator, GeneratorApply):
            instances = generator.to_instances_apply()
        else:
            instances = GeneratorApplyList(generator).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Generator:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> GeneratorList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Generator | GeneratorList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> GeneratorList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _GENERATOR_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: GeneratorFields | Sequence[GeneratorFields] = None,
        query: str | None = None,
        search_properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: GeneratorFields | Sequence[GeneratorFields] | None = None,
        query: str | None = None,
        search_property: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _GENERATOR_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: GeneratorFields,
        interval: float,
        query: str | None = None,
        search_property: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _GENERATOR_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        min_penstock: int | None = None,
        max_penstock: int | None = None,
        min_startcost: float | None = None,
        max_startcost: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> GeneratorList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            min_p_min,
            max_p_min,
            min_penstock,
            max_penstock,
            min_startcost,
            max_startcost,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    min_p_min: float | None = None,
    max_p_min: float | None = None,
    min_penstock: int | None = None,
    max_penstock: int | None = None,
    min_startcost: float | None = None,
    max_startcost: float | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if min_p_min or max_p_min:
        filters.append(dm.filters.Range(view_id.as_property_ref("pMin"), gte=min_p_min, lte=max_p_min))
    if min_penstock or max_penstock:
        filters.append(dm.filters.Range(view_id.as_property_ref("penstock"), gte=min_penstock, lte=max_penstock))
    if min_startcost or max_startcost:
        filters.append(dm.filters.Range(view_id.as_property_ref("startcost"), gte=min_startcost, lte=max_startcost))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "ProductionPlanTimeSeries",
    "ProductionPlanTimeSeriesApply",
    "ProductionPlanTimeSeriesList",
    "ProductionPlanTimeSeriesApplyList",
    "ProductionPlanTimeSeriesFields",
    "ProductionPlanTimeSeriesTextFields",
]


ProductionPlanTimeSeriesTextFields = Literal["name", "series"]
ProductionPlanTimeSeriesFields = Literal["name", "series"]

_PRODUCTIONPLANTIMESERIES_PROPERTIES_BY_FIELD = {
    "name": "name",
    "series": "series",
}


class ProductionPlanTimeSeries(DomainModel):
    space: str = "power-ops"
    name: Optional[str] = None
    series: Optional[list[str]] = None

    def as_apply(self) -> ProductionPlanTimeSeriesApply:
        return ProductionPlanTimeSeriesApply(
            external_id=self.external_id,
            name=self.name,
            series=self.series,
        )


class ProductionPlanTimeSeriesApply(DomainModelApply):
    space: str = "power-ops"
    name: Optional[str] = None
    series: Optional[list[str]] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.series is not None:
            properties["series"] = self.series
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "ProductionPlanTimeSeries"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ProductionPlanTimeSeriesList(TypeList[ProductionPlanTimeSeries]):
    _NODE = ProductionPlanTimeSeries

    def as_apply(self) -> ProductionPlanTimeSeriesApplyList:
        return ProductionPlanTimeSeriesApplyList([node.as_apply() for node in self.data])


class ProductionPlanTimeSeriesApplyList(TypeApplyList[ProductionPlanTimeSeriesApply]):
    _NODE = ProductionPlanTimeSeriesApply



from .echo import Echo
from .main import (
    DATAMODEL_ID_TO_RESYNC_NAME,
    MODELS_BY_NAME,
    V2_MODELS_BY_NAME,
    apply,
    destroy,
    init,
    migration,
    plan,
    validate,
)

__all__ = [
    "apply",
    "plan",
    "init",
    "destroy",
    "validate",
    "MODELS_BY_NAME",
    "V2_MODELS_BY_NAME",
    "Echo",
    "migration",
    "DATAMODEL_ID_TO_RESYNC_NAME",
]



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    ReserveBid,
    ReserveBidApply,
    ReserveBidApplyList,
    ReserveBidFields,
    ReserveBidList,
    ReserveBidTextFields,
)
from cognite.powerops.client._generated.data_classes._reserve_bid import _RESERVEBID_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class ReserveBidAPI(TypeAPI[ReserveBid, ReserveBidApply, ReserveBidList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ReserveBid,
            class_apply_type=ReserveBidApply,
            class_list=ReserveBidList,
        )
        self._view_id = view_id

    def apply(
        self, reserve_bid: ReserveBidApply | Sequence[ReserveBidApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(reserve_bid, ReserveBidApply):
            instances = reserve_bid.to_instances_apply()
        else:
            instances = ReserveBidApplyList(reserve_bid).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ReserveBid:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ReserveBidList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ReserveBid | ReserveBidList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: ReserveBidTextFields | Sequence[ReserveBidTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ReserveBidList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            revision_number,
            revision_number_prefix,
            type,
            type_prefix,
            process_type,
            process_type_prefix,
            sender,
            receiver,
            min_created_date_time,
            max_created_date_time,
            domain,
            subject,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _RESERVEBID_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidFields | Sequence[ReserveBidFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ReserveBidTextFields | Sequence[ReserveBidTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidFields | Sequence[ReserveBidFields] | None = None,
        group_by: ReserveBidFields | Sequence[ReserveBidFields] = None,
        query: str | None = None,
        search_properties: ReserveBidTextFields | Sequence[ReserveBidTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidFields | Sequence[ReserveBidFields] | None = None,
        group_by: ReserveBidFields | Sequence[ReserveBidFields] | None = None,
        query: str | None = None,
        search_property: ReserveBidTextFields | Sequence[ReserveBidTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            revision_number,
            revision_number_prefix,
            type,
            type_prefix,
            process_type,
            process_type_prefix,
            sender,
            receiver,
            min_created_date_time,
            max_created_date_time,
            domain,
            subject,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _RESERVEBID_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ReserveBidFields,
        interval: float,
        query: str | None = None,
        search_property: ReserveBidTextFields | Sequence[ReserveBidTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            revision_number,
            revision_number_prefix,
            type,
            type_prefix,
            process_type,
            process_type_prefix,
            sender,
            receiver,
            min_created_date_time,
            max_created_date_time,
            domain,
            subject,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _RESERVEBID_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        revision_number: str | list[str] | None = None,
        revision_number_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        process_type: str | list[str] | None = None,
        process_type_prefix: str | None = None,
        sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ReserveBidList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            revision_number,
            revision_number_prefix,
            type,
            type_prefix,
            process_type,
            process_type_prefix,
            sender,
            receiver,
            min_created_date_time,
            max_created_date_time,
            domain,
            subject,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    m_rid: str | list[str] | None = None,
    m_rid_prefix: str | None = None,
    revision_number: str | list[str] | None = None,
    revision_number_prefix: str | None = None,
    type: str | list[str] | None = None,
    type_prefix: str | None = None,
    process_type: str | list[str] | None = None,
    process_type_prefix: str | None = None,
    sender: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    receiver: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    min_created_date_time: datetime.datetime | None = None,
    max_created_date_time: datetime.datetime | None = None,
    domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    subject: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if m_rid and isinstance(m_rid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("mRID"), value=m_rid))
    if m_rid and isinstance(m_rid, list):
        filters.append(dm.filters.In(view_id.as_property_ref("mRID"), values=m_rid))
    if m_rid_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("mRID"), value=m_rid_prefix))
    if revision_number and isinstance(revision_number, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("revisionNumber"), value=revision_number))
    if revision_number and isinstance(revision_number, list):
        filters.append(dm.filters.In(view_id.as_property_ref("revisionNumber"), values=revision_number))
    if revision_number_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("revisionNumber"), value=revision_number_prefix))
    if type and isinstance(type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("type"), value=type))
    if type and isinstance(type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("type"), values=type))
    if type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("type"), value=type_prefix))
    if process_type and isinstance(process_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("processType"), value=process_type))
    if process_type and isinstance(process_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("processType"), values=process_type))
    if process_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("processType"), value=process_type_prefix))
    if sender and isinstance(sender, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("sender"), value={"space": "power-ops", "externalId": sender})
        )
    if sender and isinstance(sender, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("sender"), value={"space": sender[0], "externalId": sender[1]})
        )
    if sender and isinstance(sender, list) and isinstance(sender[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("sender"),
                values=[{"space": "power-ops", "externalId": item} for item in sender],
            )
        )
    if sender and isinstance(sender, list) and isinstance(sender[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("sender"), values=[{"space": item[0], "externalId": item[1]} for item in sender]
            )
        )
    if receiver and isinstance(receiver, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("receiver"), value={"space": "power-ops", "externalId": receiver})
        )
    if receiver and isinstance(receiver, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("receiver"), value={"space": receiver[0], "externalId": receiver[1]}
            )
        )
    if receiver and isinstance(receiver, list) and isinstance(receiver[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("receiver"),
                values=[{"space": "power-ops", "externalId": item} for item in receiver],
            )
        )
    if receiver and isinstance(receiver, list) and isinstance(receiver[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("receiver"),
                values=[{"space": item[0], "externalId": item[1]} for item in receiver],
            )
        )
    if min_created_date_time or max_created_date_time:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("createdDateTime"),
                gte=min_created_date_time.isoformat() if min_created_date_time else None,
                lte=max_created_date_time.isoformat() if max_created_date_time else None,
            )
        )
    if domain and isinstance(domain, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("domain"), value={"space": "power-ops", "externalId": domain})
        )
    if domain and isinstance(domain, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("domain"), value={"space": domain[0], "externalId": domain[1]})
        )
    if domain and isinstance(domain, list) and isinstance(domain[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("domain"),
                values=[{"space": "power-ops", "externalId": item} for item in domain],
            )
        )
    if domain and isinstance(domain, list) and isinstance(domain[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("domain"), values=[{"space": item[0], "externalId": item[1]} for item in domain]
            )
        )
    if subject and isinstance(subject, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("subject"), value={"space": "power-ops", "externalId": subject})
        )
    if subject and isinstance(subject, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("subject"), value={"space": subject[0], "externalId": subject[1]})
        )
    if subject and isinstance(subject, list) and isinstance(subject[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("subject"),
                values=[{"space": "power-ops", "externalId": item} for item in subject],
            )
        )
    if subject and isinstance(subject, list) and isinstance(subject[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("subject"),
                values=[{"space": item[0], "externalId": item[1]} for item in subject],
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from pathlib import Path

from cognite.client.data_classes.data_modeling import DataModelId

from cognite.powerops.resync.models.base.dms_models import PowerOpsDMSModel, PowerOpsDMSSourceModel

_DMS_DIR = Path(__file__).parent


CapacitySourceModel = PowerOpsDMSSourceModel(
    container_file=_DMS_DIR / "capacityBid" / "containers.yaml",
)

CapacityModel = PowerOpsDMSModel(
    name="CapacityBid",
    description="The CapacityBid model describes the capacity bids for the aFRR and mFRR markets.",
    id_=DataModelId("power-ops", "capacityBid", "1"),
    view_file=_DMS_DIR / "capacityBid" / "views.yaml",
)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._duration import DurationApply
    from ._point import PointApply

__all__ = ["Series", "SeriesApply", "SeriesList", "SeriesApplyList", "SeriesFields"]
SeriesFields = Literal["time_interval_start", "time_interval_end"]

_SERIES_PROPERTIES_BY_FIELD = {
    "time_interval_start": "timeIntervalStart",
    "time_interval_end": "timeIntervalEnd",
}


class Series(DomainModel):
    space: str = "power-ops"
    time_interval_start: Optional[datetime.datetime] = Field(None, alias="timeIntervalStart")
    time_interval_end: Optional[datetime.datetime] = Field(None, alias="timeIntervalEnd")
    resolution: Optional[str] = None
    points: Optional[list[str]] = None

    def as_apply(self) -> SeriesApply:
        return SeriesApply(
            external_id=self.external_id,
            time_interval_start=self.time_interval_start,
            time_interval_end=self.time_interval_end,
            resolution=self.resolution,
            points=self.points,
        )


class SeriesApply(DomainModelApply):
    space: str = "power-ops"
    time_interval_start: Optional[datetime.datetime] = Field(None, alias="timeIntervalStart")
    time_interval_end: Optional[datetime.datetime] = Field(None, alias="timeIntervalEnd")
    resolution: Union[DurationApply, str, None] = Field(None, repr=False)
    points: Union[list[PointApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.time_interval_start is not None:
            properties["timeIntervalStart"] = self.time_interval_start.isoformat(timespec="milliseconds")
        if self.time_interval_end is not None:
            properties["timeIntervalEnd"] = self.time_interval_end.isoformat(timespec="milliseconds")
        if self.resolution is not None:
            properties["resolution"] = {
                "space": "power-ops",
                "externalId": self.resolution if isinstance(self.resolution, str) else self.resolution.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Series"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for point in self.points or []:
            edge = self._create_point_edge(point)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(point, DomainModelApply):
                instances = point._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.resolution, DomainModelApply):
            instances = self.resolution._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_point_edge(self, point: Union[str, PointApply]) -> dm.EdgeApply:
        if isinstance(point, str):
            end_node_ext_id = point
        elif isinstance(point, DomainModelApply):
            end_node_ext_id = point.external_id
        else:
            raise TypeError(f"Expected str or PointApply, got {type(point)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "Series.points"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class SeriesList(TypeList[Series]):
    _NODE = Series

    def as_apply(self) -> SeriesApplyList:
        return SeriesApplyList([node.as_apply() for node in self.data])


class SeriesApplyList(TypeApplyList[SeriesApply]):
    _NODE = SeriesApply



from __future__ import annotations

import datetime
import warnings
from collections import defaultdict
from collections.abc import Sequence
from typing import Literal, overload

import pandas as pd
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import Datapoints, DatapointsArrayList, DatapointsList, TimeSeriesList
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList
from cognite.client.data_classes.datapoints import Aggregate

from cognite.powerops.client._generated.data_classes import (
    InputTimeSeriesMapping,
    InputTimeSeriesMappingApply,
    InputTimeSeriesMappingApplyList,
    InputTimeSeriesMappingFields,
    InputTimeSeriesMappingList,
    InputTimeSeriesMappingTextFields,
)
from cognite.powerops.client._generated.data_classes._input_time_series_mapping import (
    _INPUTTIMESERIESMAPPING_PROPERTIES_BY_FIELD,
)

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, INSTANCE_QUERY_LIMIT, Aggregations, TypeAPI

ColumnNames = Literal[
    "shopObjectType", "shopObjectName", "shopAttributeName", "cdfTimeSeries", "retrieve", "aggregation"
]


class InputTimeSeriesMappingCdfTimeSeriesQuery:
    def __init__(
        self,
        client: CogniteClient,
        view_id: dm.ViewId,
        timeseries_limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ):
        self._client = client
        self._view_id = view_id
        self._timeseries_limit = timeseries_limit
        self._filter = filter

    def retrieve(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsList([])

    def retrieve_arrays(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
    ) -> DatapointsArrayList:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_arrays(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
            )
        else:
            return DatapointsArrayList([])

    def retrieve_dataframe(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | list[Aggregate] | None = None,
        granularity: str | None = None,
        limit: int | None = None,
        include_outside_points: bool = False,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "cdfTimeSeries",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                limit=limit,
                include_outside_points=include_outside_points,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_dataframe_in_tz(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "cdfTimeSeries",
    ) -> pd.DataFrame:
        external_ids = self._retrieve_timeseries_external_ids_with_extra(column_names)
        if external_ids:
            df = self._client.time_series.data.retrieve_dataframe_in_tz(
                external_id=list(external_ids),
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
            )
            is_aggregate = aggregates is not None
            return self._rename_columns(
                external_ids,
                df,
                column_names,
                is_aggregate and include_aggregate_name,
                is_aggregate and include_granularity_name,
            )
        else:
            return pd.DataFrame()

    def retrieve_latest(
        self,
        before: None | int | str | datetime.datetime = None,
    ) -> Datapoints | DatapointsList | None:
        external_ids = self._retrieve_timeseries_external_ids_with_extra()
        if external_ids:
            return self._client.time_series.data.retrieve_latest(
                external_id=list(external_ids),
                before=before,
            )
        else:
            return None

    def plot(
        self,
        start: int | str | datetime.datetime | None = None,
        end: int | str | datetime.datetime | None = None,
        *,
        aggregates: Aggregate | Sequence[Aggregate] | None = None,
        granularity: str | None = None,
        uniform_index: bool = False,
        include_aggregate_name: bool = True,
        include_granularity_name: bool = False,
        column_names: ColumnNames | list[ColumnNames] = "cdfTimeSeries",
        warning: bool = True,
        **kwargs,
    ) -> None:
        if warning:
            warnings.warn(
                "This methods if an experiment and might be removed in the future without notice.", stacklevel=2
            )
        if all(isinstance(time, datetime.datetime) and time.tzinfo is not None for time in [start, end]):
            df = self.retrieve_dataframe_in_tz(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        else:
            df = self.retrieve_dataframe(
                start=start,
                end=end,
                aggregates=aggregates,
                granularity=granularity,
                uniform_index=uniform_index,
                include_aggregate_name=include_aggregate_name,
                include_granularity_name=include_granularity_name,
                column_names=column_names,
            )
        df.plot(**kwargs)

    def _retrieve_timeseries_external_ids_with_extra(
        self, extra_properties: ColumnNames | list[ColumnNames] = "cdfTimeSeries"
    ) -> dict[str, list[str]]:
        return _retrieve_timeseries_external_ids_with_extra_cdf_time_series(
            self._client,
            self._view_id,
            self._filter,
            self._timeseries_limit,
            extra_properties,
        )

    @staticmethod
    def _rename_columns(
        external_ids: dict[str, list[str]],
        df: pd.DataFrame,
        column_names: ColumnNames | list[ColumnNames],
        include_aggregate_name: bool,
        include_granularity_name: bool,
    ) -> pd.DataFrame:
        if isinstance(column_names, str) and column_names == "cdfTimeSeries":
            return df
        splits = sum(included for included in [include_aggregate_name, include_granularity_name])
        if splits == 0:
            df.columns = ["-".join(external_ids[external_id]) for external_id in df.columns]
        else:
            column_parts = (col.rsplit("|", maxsplit=splits) for col in df.columns)
            df.columns = [
                "-".join(external_ids[external_id]) + "|" + "|".join(parts) for external_id, *parts in column_parts
            ]
        return df


class InputTimeSeriesMappingCdfTimeSeriesAPI:
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        self._client = client
        self._view_id = view_id

    def __call__(
        self,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InputTimeSeriesMappingCdfTimeSeriesQuery:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )

        return InputTimeSeriesMappingCdfTimeSeriesQuery(
            client=self._client,
            view_id=self._view_id,
            timeseries_limit=limit,
            filter=filter_,
        )

    def list(
        self,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> TimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )
        external_ids = _retrieve_timeseries_external_ids_with_extra_cdf_time_series(
            self._client, self._view_id, filter_, limit
        )
        if external_ids:
            return self._client.time_series.retrieve_multiple(external_ids=list(external_ids))
        else:
            return TimeSeriesList([])


def _retrieve_timeseries_external_ids_with_extra_cdf_time_series(
    client: CogniteClient,
    view_id: dm.ViewId,
    filter_: dm.Filter | None,
    limit: int,
    extra_properties: ColumnNames | list[ColumnNames] = "cdfTimeSeries",
) -> dict[str, list[str]]:
    properties = ["cdfTimeSeries"]
    if extra_properties == "cdfTimeSeries":
        ...
    elif isinstance(extra_properties, str) and extra_properties != "cdfTimeSeries":
        properties.append(extra_properties)
    elif isinstance(extra_properties, list):
        properties.extend([prop for prop in extra_properties if prop != "cdfTimeSeries"])
    else:
        raise ValueError(f"Invalid value for extra_properties: {extra_properties}")

    if isinstance(extra_properties, str):
        extra_list = [extra_properties]
    else:
        extra_list = extra_properties
    has_data = dm.filters.HasData([dm.ContainerId("power-ops", "InputTimeSeriesMapping")], [view_id])
    filter_ = dm.filters.And(filter_, has_data) if filter_ else has_data

    cursor = None
    external_ids: dict[str, list[str]] = {}
    total_retrieved = 0
    while True:
        query_limit = min(INSTANCE_QUERY_LIMIT, limit - total_retrieved)
        selected_nodes = dm.query.NodeResultSetExpression(filter=filter_, limit=query_limit)
        query = dm.query.Query(
            with_={
                "nodes": selected_nodes,
            },
            select={
                "nodes": dm.query.Select(
                    [dm.query.SourceSelector(view_id, properties)],
                )
            },
            cursors={"nodes": cursor},
        )
        result = client.data_modeling.instances.query(query)
        batch_external_ids = {
            node.properties[view_id]["cdfTimeSeries"]: [node.properties[view_id].get(prop, "") for prop in extra_list]
            for node in result.data["nodes"].data
        }
        total_retrieved += len(batch_external_ids)
        external_ids.update(batch_external_ids)
        cursor = result.cursors["nodes"]
        if total_retrieved >= limit or cursor is None:
            break
    return external_ids


class InputTimeSeriesMappingTransformationsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="power-ops") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "InputTimeSeriesMapping.transformations"},
        )
        if isinstance(external_id, str):
            is_input_time_series_mapping = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_input_time_series_mapping)
            )

        else:
            is_input_time_series_mappings = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_input_time_series_mappings)
            )

    def list(
        self, input_time_series_mapping_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="power-ops"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "InputTimeSeriesMapping.transformations"},
        )
        filters.append(is_edge_type)
        if input_time_series_mapping_id:
            input_time_series_mapping_ids = (
                [input_time_series_mapping_id]
                if isinstance(input_time_series_mapping_id, str)
                else input_time_series_mapping_id
            )
            is_input_time_series_mappings = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in input_time_series_mapping_ids],
            )
            filters.append(is_input_time_series_mappings)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class InputTimeSeriesMappingAPI(
    TypeAPI[InputTimeSeriesMapping, InputTimeSeriesMappingApply, InputTimeSeriesMappingList]
):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=InputTimeSeriesMapping,
            class_apply_type=InputTimeSeriesMappingApply,
            class_list=InputTimeSeriesMappingList,
        )
        self._view_id = view_id
        self.transformations = InputTimeSeriesMappingTransformationsAPI(client)
        self.cdf_time_series = InputTimeSeriesMappingCdfTimeSeriesAPI(client, view_id)

    def apply(
        self,
        input_time_series_mapping: InputTimeSeriesMappingApply | Sequence[InputTimeSeriesMappingApply],
        replace: bool = False,
    ) -> dm.InstancesApplyResult:
        if isinstance(input_time_series_mapping, InputTimeSeriesMappingApply):
            instances = input_time_series_mapping.to_instances_apply()
        else:
            instances = InputTimeSeriesMappingApplyList(input_time_series_mapping).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> InputTimeSeriesMapping:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> InputTimeSeriesMappingList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> InputTimeSeriesMapping | InputTimeSeriesMappingList:
        if isinstance(external_id, str):
            input_time_series_mapping = self._retrieve((self._sources.space, external_id))

            transformation_edges = self.transformations.retrieve(external_id)
            input_time_series_mapping.transformations = [edge.end_node.external_id for edge in transformation_edges]

            return input_time_series_mapping
        else:
            input_time_series_mappings = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            transformation_edges = self.transformations.retrieve(external_id)
            self._set_transformations(input_time_series_mappings, transformation_edges)

            return input_time_series_mappings

    def search(
        self,
        query: str,
        properties: InputTimeSeriesMappingTextFields | Sequence[InputTimeSeriesMappingTextFields] | None = None,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InputTimeSeriesMappingList:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(
            self._view_id, query, _INPUTTIMESERIESMAPPING_PROPERTIES_BY_FIELD, properties, filter_, limit
        )

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: InputTimeSeriesMappingFields | Sequence[InputTimeSeriesMappingFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: InputTimeSeriesMappingTextFields | Sequence[InputTimeSeriesMappingTextFields] | None = None,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: InputTimeSeriesMappingFields | Sequence[InputTimeSeriesMappingFields] | None = None,
        group_by: InputTimeSeriesMappingFields | Sequence[InputTimeSeriesMappingFields] = None,
        query: str | None = None,
        search_properties: InputTimeSeriesMappingTextFields | Sequence[InputTimeSeriesMappingTextFields] | None = None,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: InputTimeSeriesMappingFields | Sequence[InputTimeSeriesMappingFields] | None = None,
        group_by: InputTimeSeriesMappingFields | Sequence[InputTimeSeriesMappingFields] | None = None,
        query: str | None = None,
        search_property: InputTimeSeriesMappingTextFields | Sequence[InputTimeSeriesMappingTextFields] | None = None,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _INPUTTIMESERIESMAPPING_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: InputTimeSeriesMappingFields,
        interval: float,
        query: str | None = None,
        search_property: InputTimeSeriesMappingTextFields | Sequence[InputTimeSeriesMappingTextFields] | None = None,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _INPUTTIMESERIESMAPPING_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_object_name: str | list[str] | None = None,
        shop_object_name_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        retrieve: str | list[str] | None = None,
        retrieve_prefix: str | None = None,
        aggregation: str | list[str] | None = None,
        aggregation_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> InputTimeSeriesMappingList:
        filter_ = _create_filter(
            self._view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_object_name,
            shop_object_name_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            retrieve,
            retrieve_prefix,
            aggregation,
            aggregation_prefix,
            external_id_prefix,
            filter,
        )

        input_time_series_mappings = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := input_time_series_mappings.as_external_ids()) > IN_FILTER_LIMIT:
                transformation_edges = self.transformations.list(limit=-1)
            else:
                transformation_edges = self.transformations.list(external_ids, limit=-1)
            self._set_transformations(input_time_series_mappings, transformation_edges)

        return input_time_series_mappings

    @staticmethod
    def _set_transformations(
        input_time_series_mappings: Sequence[InputTimeSeriesMapping], transformation_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in transformation_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for input_time_series_mapping in input_time_series_mappings:
            node_id = input_time_series_mapping.id_tuple()
            if node_id in edges_by_start_node:
                input_time_series_mapping.transformations = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]


def _create_filter(
    view_id: dm.ViewId,
    shop_object_type: str | list[str] | None = None,
    shop_object_type_prefix: str | None = None,
    shop_object_name: str | list[str] | None = None,
    shop_object_name_prefix: str | None = None,
    shop_attribute_name: str | list[str] | None = None,
    shop_attribute_name_prefix: str | None = None,
    retrieve: str | list[str] | None = None,
    retrieve_prefix: str | None = None,
    aggregation: str | list[str] | None = None,
    aggregation_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if shop_object_type and isinstance(shop_object_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopObjectType"), value=shop_object_type))
    if shop_object_type and isinstance(shop_object_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopObjectType"), values=shop_object_type))
    if shop_object_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopObjectType"), value=shop_object_type_prefix))
    if shop_object_name and isinstance(shop_object_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopObjectName"), value=shop_object_name))
    if shop_object_name and isinstance(shop_object_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopObjectName"), values=shop_object_name))
    if shop_object_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopObjectName"), value=shop_object_name_prefix))
    if shop_attribute_name and isinstance(shop_attribute_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopAttributeName"), value=shop_attribute_name))
    if shop_attribute_name and isinstance(shop_attribute_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopAttributeName"), values=shop_attribute_name))
    if shop_attribute_name_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopAttributeName"), value=shop_attribute_name_prefix)
        )
    if retrieve and isinstance(retrieve, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("retrieve"), value=retrieve))
    if retrieve and isinstance(retrieve, list):
        filters.append(dm.filters.In(view_id.as_property_ref("retrieve"), values=retrieve))
    if retrieve_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("retrieve"), value=retrieve_prefix))
    if aggregation and isinstance(aggregation, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("aggregation"), value=aggregation))
    if aggregation and isinstance(aggregation, list):
        filters.append(dm.filters.In(view_id.as_property_ref("aggregation"), values=aggregation))
    if aggregation_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("aggregation"), value=aggregation_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._date_time_interval import DateTimeIntervalApply
    from ._duration import DurationApply
    from ._market_agreement import MarketAgreementApply
    from ._market_participant import MarketParticipantApply
    from ._mba_domain import MBADomainApply
    from ._reason import ReasonApply
    from ._reserve_bid import ReserveBidApply

__all__ = [
    "BidTimeSeries",
    "BidTimeSeriesApply",
    "BidTimeSeriesList",
    "BidTimeSeriesApplyList",
    "BidTimeSeriesFields",
    "BidTimeSeriesTextFields",
]


BidTimeSeriesTextFields = Literal[
    "m_rid",
    "auction",
    "quantity_measure_unit_name",
    "currency_unit_name",
    "price_measure_unit_name",
    "registered_resources_mrid",
    "flow_direction",
    "energy_price_measure_unit",
    "standard_market_product_type",
    "original_market_product_type",
]
BidTimeSeriesFields = Literal[
    "m_rid",
    "auction",
    "quantity_measure_unit_name",
    "currency_unit_name",
    "price_measure_unit_name",
    "divisible",
    "block_bid",
    "status",
    "priority",
    "registered_resources_mrid",
    "flow_direction",
    "step_increment_quantity",
    "energy_price_measure_unit",
    "standard_market_product_type",
    "original_market_product_type",
]

_BIDTIMESERIES_PROPERTIES_BY_FIELD = {
    "m_rid": "mRID",
    "auction": "auction",
    "quantity_measure_unit_name": "quantityMeasureUnitName",
    "currency_unit_name": "currencyUnitName",
    "price_measure_unit_name": "priceMeasureUnitName",
    "divisible": "divisible",
    "block_bid": "blockBid",
    "status": "status",
    "priority": "priority",
    "registered_resources_mrid": "registeredResourcesMRID",
    "flow_direction": "flowDirection",
    "step_increment_quantity": "stepIncrementQuantity",
    "energy_price_measure_unit": "energyPriceMeasureUnit",
    "standard_market_product_type": "standardMarketProductType",
    "original_market_product_type": "originalMarketProductType",
}


class BidTimeSeries(DomainModel):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    auction: Optional[str] = None
    acquiring_domain: Optional[str] = Field(None, alias="acquiringDomain")
    connecting_domain: Optional[str] = Field(None, alias="connectingDomain")
    provider_market_participant: Optional[str] = Field(None, alias="providerMarketParticipant")
    quantity_measure_unit_name: Optional[str] = Field(None, alias="quantityMeasureUnitName")
    currency_unit_name: Optional[str] = Field(None, alias="currencyUnitName")
    price_measure_unit_name: Optional[str] = Field(None, alias="priceMeasureUnitName")
    divisible: Optional[bool] = None
    linked_bid: Optional[str] = Field(None, alias="linkedBid")
    multipart_bid: Optional[str] = Field(None, alias="multipartBid")
    exclusive_bid: Optional[str] = Field(None, alias="exclusiveBid")
    block_bid: Optional[bool] = Field(None, alias="blockBid")
    status: Optional[int] = None
    priority: Optional[int] = None
    registered_resources_mrid: Optional[str] = Field(None, alias="registeredResourcesMRID")
    flow_direction: Optional[str] = Field(None, alias="flowDirection")
    step_increment_quantity: Optional[float] = Field(None, alias="stepIncrementQuantity")
    energy_price_measure_unit: Optional[str] = Field(None, alias="energyPriceMeasureUnit")
    market_agreement: Optional[str] = Field(None, alias="marketAgreement")
    activation_constraint: Optional[str] = Field(None, alias="activationConstraint")
    resting_constraint: Optional[str] = Field(None, alias="restingConstraint")
    minimum_constraint: Optional[str] = Field(None, alias="minimumConstraint")
    maximum_constraint: Optional[str] = Field(None, alias="maximumConstraint")
    standard_market_product_type: Optional[str] = Field(None, alias="standardMarketProductType")
    original_market_product_type: Optional[str] = Field(None, alias="originalMarketProductType")
    validity_period: Optional[str] = Field(None, alias="validityPeriod")
    reason: Optional[str] = None

    def as_apply(self) -> BidTimeSeriesApply:
        return BidTimeSeriesApply(
            external_id=self.external_id,
            m_rid=self.m_rid,
            auction=self.auction,
            acquiring_domain=self.acquiring_domain,
            connecting_domain=self.connecting_domain,
            provider_market_participant=self.provider_market_participant,
            quantity_measure_unit_name=self.quantity_measure_unit_name,
            currency_unit_name=self.currency_unit_name,
            price_measure_unit_name=self.price_measure_unit_name,
            divisible=self.divisible,
            linked_bid=self.linked_bid,
            multipart_bid=self.multipart_bid,
            exclusive_bid=self.exclusive_bid,
            block_bid=self.block_bid,
            status=self.status,
            priority=self.priority,
            registered_resources_mrid=self.registered_resources_mrid,
            flow_direction=self.flow_direction,
            step_increment_quantity=self.step_increment_quantity,
            energy_price_measure_unit=self.energy_price_measure_unit,
            market_agreement=self.market_agreement,
            activation_constraint=self.activation_constraint,
            resting_constraint=self.resting_constraint,
            minimum_constraint=self.minimum_constraint,
            maximum_constraint=self.maximum_constraint,
            standard_market_product_type=self.standard_market_product_type,
            original_market_product_type=self.original_market_product_type,
            validity_period=self.validity_period,
            reason=self.reason,
        )


class BidTimeSeriesApply(DomainModelApply):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    auction: Optional[str] = None
    acquiring_domain: Union[MBADomainApply, str, None] = Field(None, repr=False, alias="acquiringDomain")
    connecting_domain: Union[MBADomainApply, str, None] = Field(None, repr=False, alias="connectingDomain")
    provider_market_participant: Union[MarketParticipantApply, str, None] = Field(
        None, repr=False, alias="providerMarketParticipant"
    )
    quantity_measure_unit_name: Optional[str] = Field(None, alias="quantityMeasureUnitName")
    currency_unit_name: Optional[str] = Field(None, alias="currencyUnitName")
    price_measure_unit_name: Optional[str] = Field(None, alias="priceMeasureUnitName")
    divisible: Optional[bool] = None
    linked_bid: Union[ReserveBidApply, str, None] = Field(None, repr=False, alias="linkedBid")
    multipart_bid: Union[ReserveBidApply, str, None] = Field(None, repr=False, alias="multipartBid")
    exclusive_bid: Union[ReserveBidApply, str, None] = Field(None, repr=False, alias="exclusiveBid")
    block_bid: Optional[bool] = Field(None, alias="blockBid")
    status: Optional[int] = None
    priority: Optional[int] = None
    registered_resources_mrid: Optional[str] = Field(None, alias="registeredResourcesMRID")
    flow_direction: Optional[str] = Field(None, alias="flowDirection")
    step_increment_quantity: Optional[float] = Field(None, alias="stepIncrementQuantity")
    energy_price_measure_unit: Optional[str] = Field(None, alias="energyPriceMeasureUnit")
    market_agreement: Union[MarketAgreementApply, str, None] = Field(None, repr=False, alias="marketAgreement")
    activation_constraint: Union[DurationApply, str, None] = Field(None, repr=False, alias="activationConstraint")
    resting_constraint: Union[DurationApply, str, None] = Field(None, repr=False, alias="restingConstraint")
    minimum_constraint: Union[DurationApply, str, None] = Field(None, repr=False, alias="minimumConstraint")
    maximum_constraint: Union[DurationApply, str, None] = Field(None, repr=False, alias="maximumConstraint")
    standard_market_product_type: Optional[str] = Field(None, alias="standardMarketProductType")
    original_market_product_type: Optional[str] = Field(None, alias="originalMarketProductType")
    validity_period: Union[DateTimeIntervalApply, str, None] = Field(None, repr=False, alias="validityPeriod")
    reason: Union[ReasonApply, str, None] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.m_rid is not None:
            properties["mRID"] = self.m_rid
        if self.auction is not None:
            properties["auction"] = self.auction
        if self.acquiring_domain is not None:
            properties["acquiringDomain"] = {
                "space": "power-ops",
                "externalId": self.acquiring_domain
                if isinstance(self.acquiring_domain, str)
                else self.acquiring_domain.external_id,
            }
        if self.connecting_domain is not None:
            properties["connectingDomain"] = {
                "space": "power-ops",
                "externalId": self.connecting_domain
                if isinstance(self.connecting_domain, str)
                else self.connecting_domain.external_id,
            }
        if self.provider_market_participant is not None:
            properties["providerMarketParticipant"] = {
                "space": "power-ops",
                "externalId": self.provider_market_participant
                if isinstance(self.provider_market_participant, str)
                else self.provider_market_participant.external_id,
            }
        if self.quantity_measure_unit_name is not None:
            properties["quantityMeasureUnitName"] = self.quantity_measure_unit_name
        if self.currency_unit_name is not None:
            properties["currencyUnitName"] = self.currency_unit_name
        if self.price_measure_unit_name is not None:
            properties["priceMeasureUnitName"] = self.price_measure_unit_name
        if self.divisible is not None:
            properties["divisible"] = self.divisible
        if self.linked_bid is not None:
            properties["linkedBid"] = {
                "space": "power-ops",
                "externalId": self.linked_bid if isinstance(self.linked_bid, str) else self.linked_bid.external_id,
            }
        if self.multipart_bid is not None:
            properties["multipartBid"] = {
                "space": "power-ops",
                "externalId": self.multipart_bid
                if isinstance(self.multipart_bid, str)
                else self.multipart_bid.external_id,
            }
        if self.exclusive_bid is not None:
            properties["exclusiveBid"] = {
                "space": "power-ops",
                "externalId": self.exclusive_bid
                if isinstance(self.exclusive_bid, str)
                else self.exclusive_bid.external_id,
            }
        if self.block_bid is not None:
            properties["blockBid"] = self.block_bid
        if self.status is not None:
            properties["status"] = self.status
        if self.priority is not None:
            properties["priority"] = self.priority
        if self.registered_resources_mrid is not None:
            properties["registeredResourcesMRID"] = self.registered_resources_mrid
        if self.flow_direction is not None:
            properties["flowDirection"] = self.flow_direction
        if self.step_increment_quantity is not None:
            properties["stepIncrementQuantity"] = self.step_increment_quantity
        if self.energy_price_measure_unit is not None:
            properties["energyPriceMeasureUnit"] = self.energy_price_measure_unit
        if self.market_agreement is not None:
            properties["marketAgreement"] = {
                "space": "power-ops",
                "externalId": self.market_agreement
                if isinstance(self.market_agreement, str)
                else self.market_agreement.external_id,
            }
        if self.activation_constraint is not None:
            properties["activationConstraint"] = {
                "space": "power-ops",
                "externalId": self.activation_constraint
                if isinstance(self.activation_constraint, str)
                else self.activation_constraint.external_id,
            }
        if self.resting_constraint is not None:
            properties["restingConstraint"] = {
                "space": "power-ops",
                "externalId": self.resting_constraint
                if isinstance(self.resting_constraint, str)
                else self.resting_constraint.external_id,
            }
        if self.minimum_constraint is not None:
            properties["minimumConstraint"] = {
                "space": "power-ops",
                "externalId": self.minimum_constraint
                if isinstance(self.minimum_constraint, str)
                else self.minimum_constraint.external_id,
            }
        if self.maximum_constraint is not None:
            properties["maximumConstraint"] = {
                "space": "power-ops",
                "externalId": self.maximum_constraint
                if isinstance(self.maximum_constraint, str)
                else self.maximum_constraint.external_id,
            }
        if self.standard_market_product_type is not None:
            properties["standardMarketProductType"] = self.standard_market_product_type
        if self.original_market_product_type is not None:
            properties["originalMarketProductType"] = self.original_market_product_type
        if self.validity_period is not None:
            properties["validityPeriod"] = {
                "space": "power-ops",
                "externalId": self.validity_period
                if isinstance(self.validity_period, str)
                else self.validity_period.external_id,
            }
        if self.reason is not None:
            properties["reason"] = {
                "space": "power-ops",
                "externalId": self.reason if isinstance(self.reason, str) else self.reason.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BidTimeSeries"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        if isinstance(self.acquiring_domain, DomainModelApply):
            instances = self.acquiring_domain._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.connecting_domain, DomainModelApply):
            instances = self.connecting_domain._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.provider_market_participant, DomainModelApply):
            instances = self.provider_market_participant._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.linked_bid, DomainModelApply):
            instances = self.linked_bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.multipart_bid, DomainModelApply):
            instances = self.multipart_bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.exclusive_bid, DomainModelApply):
            instances = self.exclusive_bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.market_agreement, DomainModelApply):
            instances = self.market_agreement._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.activation_constraint, DomainModelApply):
            instances = self.activation_constraint._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.resting_constraint, DomainModelApply):
            instances = self.resting_constraint._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.minimum_constraint, DomainModelApply):
            instances = self.minimum_constraint._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.maximum_constraint, DomainModelApply):
            instances = self.maximum_constraint._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.validity_period, DomainModelApply):
            instances = self.validity_period._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.reason, DomainModelApply):
            instances = self.reason._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class BidTimeSeriesList(TypeList[BidTimeSeries]):
    _NODE = BidTimeSeries

    def as_apply(self) -> BidTimeSeriesApplyList:
        return BidTimeSeriesApplyList([node.as_apply() for node in self.data])


class BidTimeSeriesApplyList(TypeApplyList[BidTimeSeriesApply]):
    _NODE = BidTimeSeriesApply



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    ScenarioTemplate,
    ScenarioTemplateApply,
    ScenarioTemplateApplyList,
    ScenarioTemplateFields,
    ScenarioTemplateList,
    ScenarioTemplateTextFields,
)
from cognite.powerops.client._generated.data_classes._scenario_template import _SCENARIOTEMPLATE_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class ScenarioTemplateAPI(TypeAPI[ScenarioTemplate, ScenarioTemplateApply, ScenarioTemplateList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ScenarioTemplate,
            class_apply_type=ScenarioTemplateApply,
            class_list=ScenarioTemplateList,
        )
        self._view_id = view_id

    def apply(
        self, scenario_template: ScenarioTemplateApply | Sequence[ScenarioTemplateApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(scenario_template, ScenarioTemplateApply):
            instances = scenario_template.to_instances_apply()
        else:
            instances = ScenarioTemplateApplyList(scenario_template).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ScenarioTemplate:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioTemplateList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ScenarioTemplate | ScenarioTemplateList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: ScenarioTemplateTextFields | Sequence[ScenarioTemplateTextFields] | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioTemplateList:
        filter_ = _create_filter(
            self._view_id,
            watercourse,
            watercourse_prefix,
            shop_version,
            shop_version_prefix,
            template_version,
            template_version_prefix,
            base_mapping,
            output_definitions,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _SCENARIOTEMPLATE_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioTemplateFields | Sequence[ScenarioTemplateFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ScenarioTemplateTextFields | Sequence[ScenarioTemplateTextFields] | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioTemplateFields | Sequence[ScenarioTemplateFields] | None = None,
        group_by: ScenarioTemplateFields | Sequence[ScenarioTemplateFields] = None,
        query: str | None = None,
        search_properties: ScenarioTemplateTextFields | Sequence[ScenarioTemplateTextFields] | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioTemplateFields | Sequence[ScenarioTemplateFields] | None = None,
        group_by: ScenarioTemplateFields | Sequence[ScenarioTemplateFields] | None = None,
        query: str | None = None,
        search_property: ScenarioTemplateTextFields | Sequence[ScenarioTemplateTextFields] | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            watercourse,
            watercourse_prefix,
            shop_version,
            shop_version_prefix,
            template_version,
            template_version_prefix,
            base_mapping,
            output_definitions,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SCENARIOTEMPLATE_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ScenarioTemplateFields,
        interval: float,
        query: str | None = None,
        search_property: ScenarioTemplateTextFields | Sequence[ScenarioTemplateTextFields] | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            watercourse,
            watercourse_prefix,
            shop_version,
            shop_version_prefix,
            template_version,
            template_version_prefix,
            base_mapping,
            output_definitions,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SCENARIOTEMPLATE_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioTemplateList:
        filter_ = _create_filter(
            self._view_id,
            watercourse,
            watercourse_prefix,
            shop_version,
            shop_version_prefix,
            template_version,
            template_version_prefix,
            base_mapping,
            output_definitions,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    shop_version: str | list[str] | None = None,
    shop_version_prefix: str | None = None,
    template_version: str | list[str] | None = None,
    template_version_prefix: str | None = None,
    base_mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    output_definitions: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if shop_version and isinstance(shop_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopVersion"), value=shop_version))
    if shop_version and isinstance(shop_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopVersion"), values=shop_version))
    if shop_version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopVersion"), value=shop_version_prefix))
    if template_version and isinstance(template_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("templateVersion"), value=template_version))
    if template_version and isinstance(template_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("templateVersion"), values=template_version))
    if template_version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("templateVersion"), value=template_version_prefix))
    if base_mapping and isinstance(base_mapping, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("baseMapping"), value={"space": "power-ops", "externalId": base_mapping}
            )
        )
    if base_mapping and isinstance(base_mapping, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("baseMapping"), value={"space": base_mapping[0], "externalId": base_mapping[1]}
            )
        )
    if base_mapping and isinstance(base_mapping, list) and isinstance(base_mapping[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("baseMapping"),
                values=[{"space": "power-ops", "externalId": item} for item in base_mapping],
            )
        )
    if base_mapping and isinstance(base_mapping, list) and isinstance(base_mapping[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("baseMapping"),
                values=[{"space": item[0], "externalId": item[1]} for item in base_mapping],
            )
        )
    if output_definitions and isinstance(output_definitions, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("outputDefinitions"),
                value={"space": "power-ops", "externalId": output_definitions},
            )
        )
    if output_definitions and isinstance(output_definitions, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("outputDefinitions"),
                value={"space": output_definitions[0], "externalId": output_definitions[1]},
            )
        )
    if output_definitions and isinstance(output_definitions, list) and isinstance(output_definitions[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("outputDefinitions"),
                values=[{"space": "power-ops", "externalId": item} for item in output_definitions],
            )
        )
    if output_definitions and isinstance(output_definitions, list) and isinstance(output_definitions[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("outputDefinitions"),
                values=[{"space": item[0], "externalId": item[1]} for item in output_definitions],
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._date_transformation import DateTransformationApply
    from ._nord_pool_market import NordPoolMarketApply

__all__ = [
    "BenchmarkBid",
    "BenchmarkBidApply",
    "BenchmarkBidList",
    "BenchmarkBidApplyList",
    "BenchmarkBidFields",
    "BenchmarkBidTextFields",
]


BenchmarkBidTextFields = Literal["name"]
BenchmarkBidFields = Literal["name"]

_BENCHMARKBID_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class BenchmarkBid(DomainModel):
    space: str = "power-ops"
    name: Optional[str] = None
    market: Optional[str] = None
    date: Optional[list[str]] = None

    def as_apply(self) -> BenchmarkBidApply:
        return BenchmarkBidApply(
            external_id=self.external_id,
            name=self.name,
            market=self.market,
            date=self.date,
        )


class BenchmarkBidApply(DomainModelApply):
    space: str = "power-ops"
    name: Optional[str] = None
    market: Union[NordPoolMarketApply, str, None] = Field(None, repr=False)
    date: Union[list[DateTransformationApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.market is not None:
            properties["market"] = {
                "space": "power-ops",
                "externalId": self.market if isinstance(self.market, str) else self.market.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BenchmarkBid"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for date in self.date or []:
            edge = self._create_date_edge(date)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(date, DomainModelApply):
                instances = date._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.market, DomainModelApply):
            instances = self.market._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_date_edge(self, date: Union[str, DateTransformationApply]) -> dm.EdgeApply:
        if isinstance(date, str):
            end_node_ext_id = date
        elif isinstance(date, DomainModelApply):
            end_node_ext_id = date.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(date)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "BenchmarkBid.date"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class BenchmarkBidList(TypeList[BenchmarkBid]):
    _NODE = BenchmarkBid

    def as_apply(self) -> BenchmarkBidApplyList:
        return BenchmarkBidApplyList([node.as_apply() for node in self.data])


class BenchmarkBidApplyList(TypeApplyList[BenchmarkBidApply]):
    _NODE = BenchmarkBidApply



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    ScenarioMapping,
    ScenarioMappingApply,
    ScenarioMappingApplyList,
    ScenarioMappingFields,
    ScenarioMappingList,
    ScenarioMappingTextFields,
)
from cognite.powerops.client._generated.data_classes._scenario_mapping import _SCENARIOMAPPING_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, Aggregations, TypeAPI


class ScenarioMappingMappingOverrideAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="power-ops") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ScenarioMapping.mappingOverride"},
        )
        if isinstance(external_id, str):
            is_scenario_mapping = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_scenario_mapping)
            )

        else:
            is_scenario_mappings = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_scenario_mappings)
            )

    def list(
        self, scenario_mapping_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="power-ops"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ScenarioMapping.mappingOverride"},
        )
        filters.append(is_edge_type)
        if scenario_mapping_id:
            scenario_mapping_ids = (
                [scenario_mapping_id] if isinstance(scenario_mapping_id, str) else scenario_mapping_id
            )
            is_scenario_mappings = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in scenario_mapping_ids],
            )
            filters.append(is_scenario_mappings)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ScenarioMappingAPI(TypeAPI[ScenarioMapping, ScenarioMappingApply, ScenarioMappingList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ScenarioMapping,
            class_apply_type=ScenarioMappingApply,
            class_list=ScenarioMappingList,
        )
        self._view_id = view_id
        self.mapping_override = ScenarioMappingMappingOverrideAPI(client)

    def apply(
        self, scenario_mapping: ScenarioMappingApply | Sequence[ScenarioMappingApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(scenario_mapping, ScenarioMappingApply):
            instances = scenario_mapping.to_instances_apply()
        else:
            instances = ScenarioMappingApplyList(scenario_mapping).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ScenarioMapping:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioMappingList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ScenarioMapping | ScenarioMappingList:
        if isinstance(external_id, str):
            scenario_mapping = self._retrieve((self._sources.space, external_id))

            mapping_override_edges = self.mapping_override.retrieve(external_id)
            scenario_mapping.mapping_override = [edge.end_node.external_id for edge in mapping_override_edges]

            return scenario_mapping
        else:
            scenario_mappings = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            mapping_override_edges = self.mapping_override.retrieve(external_id)
            self._set_mapping_override(scenario_mappings, mapping_override_edges)

            return scenario_mappings

    def search(
        self,
        query: str,
        properties: ScenarioMappingTextFields | Sequence[ScenarioMappingTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioMappingList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _SCENARIOMAPPING_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioMappingFields | Sequence[ScenarioMappingFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ScenarioMappingTextFields | Sequence[ScenarioMappingTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioMappingFields | Sequence[ScenarioMappingFields] | None = None,
        group_by: ScenarioMappingFields | Sequence[ScenarioMappingFields] = None,
        query: str | None = None,
        search_properties: ScenarioMappingTextFields | Sequence[ScenarioMappingTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioMappingFields | Sequence[ScenarioMappingFields] | None = None,
        group_by: ScenarioMappingFields | Sequence[ScenarioMappingFields] | None = None,
        query: str | None = None,
        search_property: ScenarioMappingTextFields | Sequence[ScenarioMappingTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SCENARIOMAPPING_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ScenarioMappingFields,
        interval: float,
        query: str | None = None,
        search_property: ScenarioMappingTextFields | Sequence[ScenarioMappingTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SCENARIOMAPPING_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ScenarioMappingList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )

        scenario_mappings = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := scenario_mappings.as_external_ids()) > IN_FILTER_LIMIT:
                mapping_override_edges = self.mapping_override.list(limit=-1)
            else:
                mapping_override_edges = self.mapping_override.list(external_ids, limit=-1)
            self._set_mapping_override(scenario_mappings, mapping_override_edges)

        return scenario_mappings

    @staticmethod
    def _set_mapping_override(scenario_mappings: Sequence[ScenarioMapping], mapping_override_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in mapping_override_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for scenario_mapping in scenario_mappings:
            node_id = scenario_mapping.id_tuple()
            if node_id in edges_by_start_node:
                scenario_mapping.mapping_override = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    shop_type: str | list[str] | None = None,
    shop_type_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if shop_type and isinstance(shop_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopType"), value=shop_type))
    if shop_type and isinstance(shop_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopType"), values=shop_type))
    if shop_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopType"), value=shop_type_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
